@@ -12,6 +12,7 @@ from awkward._backends.backend import Backend
 from awkward._layout import maybe_posaxis
 from awkward._meta.unionmeta import UnionMeta
 from awkward._nplikes.array_like import ArrayLike
+from awkward._nplikes.cupy import Cupy
 from awkward._nplikes.numpy import Numpy
 from awkward._nplikes.numpy_like import IndexType, NumpyMetadata
 from awkward._nplikes.placeholder import PlaceholderArray
@@ -38,7 +39,7 @@ from awkward.contents.content import (
     ToArrowOptions,
 )
 from awkward.errors import AxisError
-from awkward.forms.form import Form
+from awkward.forms.form import Form, FormKeyPathT
 from awkward.forms.unionform import UnionForm
 from awkward.index import Index, Index8, Index64
 
@@ -47,6 +48,7 @@ if TYPE_CHECKING:
 
 np = NumpyMetadata.instance()
 numpy = Numpy.instance()
+MAX_UNION_CONTENTS = 2**7  # We use int8 tags, 0-127
 
 
 @final
@@ -229,6 +231,10 @@ class UnionArray(UnionMeta[Content], Content):
         parameters=None,
         mergebool=False,
     ):
+        # Note: to help merge more than 128 arrays, tags *can* have type ak.index.Index64.
+        # This is only supported when index is also Index64,
+        # and all indexed contents are also Index64.
+        # We still require that this reduces to no more than 128 variants.
         self_index = index
         self_tags = tags
         self_contents = contents
@@ -298,6 +304,10 @@ class UnionArray(UnionMeta[Content], Content):
 
                     # Did we fail to merge any of the final outer contents with this inner union content?
                     if unmerged:
+                        if len(contents) >= MAX_UNION_CONTENTS:
+                            raise ValueError(
+                                "UnionArray does not support more than 128 content types"
+                            )
                         backend.maybe_kernel_error(
                             backend[
                                 "awkward_UnionArray_simplify",
@@ -372,6 +382,10 @@ class UnionArray(UnionMeta[Content], Content):
                         break
 
                 if unmerged:
+                    if len(contents) >= MAX_UNION_CONTENTS:
+                        raise ValueError(
+                            "UnionArray does not support more than 128 content types"
+                        )
                     backend.maybe_kernel_error(
                         backend[
                             "awkward_UnionArray_simplify_one",
@@ -392,11 +406,6 @@ class UnionArray(UnionMeta[Content], Content):
                     )
                     contents.append(self_cont)
 
-        if len(contents) > 2**7:
-            raise NotImplementedError(
-                "FIXME: handle UnionArray with more than 127 contents"
-            )
-
         # If any contents are non-categorical index types, we can merge them into the union
         # This is safe, because any remaining index types at this point in the routine are not considered
         # mergeable with the other contents. This means none of the other contents are option or index types,
@@ -451,6 +460,15 @@ class UnionArray(UnionMeta[Content], Content):
             form_key=form_key,
         )
 
+    def _form_with_key_path(self, path: FormKeyPathT) -> UnionForm:
+        return self.form_cls(
+            self._tags.form,
+            self._index.form,
+            [x._form_with_key_path((*path, i)) for i, x in enumerate(self._contents)],
+            parameters=self._parameters,
+            form_key=repr(path),
+        )
+
     def _to_buffers(
         self,
         form: Form,
@@ -527,11 +545,11 @@ class UnionArray(UnionMeta[Content], Content):
         if isinstance(self._tags, PlaceholderArray) or isinstance(
             self._index, PlaceholderArray
         ):
-            return False
+            return True
         for content in self._contents:
             if content._is_getitem_at_placeholder():
-                return False
-        return True
+                return True
+        return False
 
     def _getitem_at(self, where: IndexType):
         if not self._backend.nplike.known_data:
@@ -878,11 +896,13 @@ class UnionArray(UnionMeta[Content], Content):
             )
             contents = []
 
+            keep_offsets = []
             for i in range(len(self._contents)):
                 offsets, flattened = self._contents[i]._offsets_and_flattened(
                     axis, depth
                 )
                 offsetsraws[i] = offsets.ptr
+                keep_offsets.append(offsets)
                 contents.append(flattened)
                 has_offsets = offsets.length != 0
 
@@ -895,23 +915,40 @@ class UnionArray(UnionMeta[Content], Content):
                     and self._tags.nplike is self._backend.index_nplike
                     and self._index.nplike is self._backend.index_nplike
                 )
-                self._backend.maybe_kernel_error(
-                    self._backend[
-                        "awkward_UnionArray_flatten_length",
-                        total_length.dtype.type,
-                        self._tags.dtype.type,
-                        self._index.dtype.type,
-                        np.int64,
-                    ](
-                        total_length.data,
-                        self._tags.data,
-                        self._index.data,
-                        self._tags.length,
-                        offsetsraws.ctypes.data_as(
-                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
-                        ),
+                if self._backend.nplike == Numpy.instance():
+                    self._backend.maybe_kernel_error(
+                        self._backend[
+                            "awkward_UnionArray_flatten_length",
+                            total_length.dtype.type,
+                            self._tags.dtype.type,
+                            self._index.dtype.type,
+                            np.int64,
+                        ](
+                            total_length.data,
+                            self._tags.data,
+                            self._index.data,
+                            self._tags.length,
+                            offsetsraws.ctypes.data_as(
+                                ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
+                            ),
+                        )
+                    )
+                elif self._backend.nplike == Cupy.instance():
+                    self._backend.maybe_kernel_error(
+                        self._backend[
+                            "awkward_UnionArray_flatten_length",
+                            total_length.dtype.type,
+                            self._tags.dtype.type,
+                            self._index.dtype.type,
+                            np.int64,
+                        ](
+                            total_length.data,
+                            self._tags.data,
+                            self._index.data,
+                            self._tags.length,
+                            offsetsraws,
+                        )
                     )
-                )
 
                 totags = ak.index.Index8.empty(
                     total_length[0], nplike=self._backend.index_nplike
@@ -930,28 +967,48 @@ class UnionArray(UnionMeta[Content], Content):
                     and self._tags.nplike is self._backend.index_nplike
                     and self._index.nplike is self._backend.index_nplike
                 )
-                self._backend.maybe_kernel_error(
-                    self._backend[
-                        "awkward_UnionArray_flatten_combine",
-                        totags.dtype.type,
-                        toindex.dtype.type,
-                        tooffsets.dtype.type,
-                        self._tags.dtype.type,
-                        self._index.dtype.type,
-                        np.int64,
-                    ](
-                        totags.data,
-                        toindex.data,
-                        tooffsets.data,
-                        self._tags.data,
-                        self._index.data,
-                        self._tags.length,
-                        offsetsraws.ctypes.data_as(
-                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
-                        ),
+                if self._backend.nplike == Numpy.instance():
+                    self._backend.maybe_kernel_error(
+                        self._backend[
+                            "awkward_UnionArray_flatten_combine",
+                            totags.dtype.type,
+                            toindex.dtype.type,
+                            tooffsets.dtype.type,
+                            self._tags.dtype.type,
+                            self._index.dtype.type,
+                            np.int64,
+                        ](
+                            totags.data,
+                            toindex.data,
+                            tooffsets.data,
+                            self._tags.data,
+                            self._index.data,
+                            self._tags.length,
+                            offsetsraws.ctypes.data_as(
+                                ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
+                            ),
+                        )
+                    )
+                elif self._backend.nplike == Cupy.instance():
+                    self._backend.maybe_kernel_error(
+                        self._backend[
+                            "awkward_UnionArray_flatten_combine",
+                            totags.dtype.type,
+                            toindex.dtype.type,
+                            tooffsets.dtype.type,
+                            self._tags.dtype.type,
+                            self._index.dtype.type,
+                            np.int64,
+                        ](
+                            totags.data,
+                            toindex.data,
+                            tooffsets.data,
+                            self._tags.data,
+                            self._index.data,
+                            self._tags.length,
+                            offsetsraws,
+                        )
                     )
-                )
-
                 return (
                     tooffsets,
                     UnionArray(
@@ -1067,8 +1124,8 @@ class UnionArray(UnionMeta[Content], Content):
             )
         )
 
-        if len(contents) > 2**7:
-            raise AssertionError("FIXME: handle UnionArray with more than 127 contents")
+        if len(contents) > MAX_UNION_CONTENTS:
+            raise ValueError("UnionArray cannot have more than 128 content types")
 
         return ak.contents.UnionArray.simplified(
             tags, index, contents, parameters=self._parameters
@@ -1196,8 +1253,8 @@ class UnionArray(UnionMeta[Content], Content):
 
                 nextcontents.append(array)
 
-        if len(nextcontents) > 127:
-            raise ValueError("FIXME: handle UnionArray with more than 127 contents")
+        if len(nextcontents) > MAX_UNION_CONTENTS:
+            raise ValueError("UnionArray cannot have more than 128 content types")
 
         next = ak.contents.UnionArray.simplified(
             nexttags,
@@ -1479,7 +1536,8 @@ class UnionArray(UnionMeta[Content], Content):
         types = pyarrow.union(
             [
                 pyarrow.field(str(i), values[i].type).with_nullable(
-                    mask_node is not None or self._contents[i].is_option
+                    mask_node is not None
+                    or self._contents[i]._arrow_needs_option_type()
                 )
                 for i in range(len(values))
             ],
